{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE TypeSynonymInstances #-}
{-# LANGUAGE OverloadedStrings #-}

module Eval where

import Text.Parsec.Pos
import Data.Bits
import qualified Data.Map.Strict as M
import System.IO
import Control.Monad.Trans
import Control.Monad.State.Lazy
import Control.Monad.Except
import Control.Monad.Trans.Except

import Types

import Data.Aeson

type Stack a = [a]

type ReturnStack = Stack Int
type RunStack    = Stack Literal
type Locals      = Stack Literal
type LocalsStack = Stack Locals
type Args        = Stack Literal
type ArgsStack   = Stack Args

data ST = ST { program    :: [(Command,Line)],
               lmap       :: LabelMap,

               retstack   :: ReturnStack,
               runstack   :: RunStack,

               locals      :: Locals,
               localsstack :: LocalsStack,

               calleeargs  :: Args,
               callerargs  :: Args,
               argsstack  :: ArgsStack,

               returnReg   :: Maybe Literal,

               instCount  :: Int,

               outputs :: [String],
               inputs :: Bool
             }
    deriving Show

runstackJSON :: ST -> Line -> Value
runstackJSON st l = object [
    "runstack" .= (toJSON $ runstack st),
    "returnstack" .= (toJSON $ retstack st),
    "instCount" .= (toJSON $ instCount st),
    "input" .= (toJSON $ inputs st),
    "output" .= toJSON (reverse $ outputs st),
    "pc"          .= toJSON l
  ]

type ExSTM = StateT ST (ExceptT String IO)

-- Functions for working within the ST transformer

-- input :: ExSTM Int
-- input = liftIO (readLn :: IO Int)

input :: ExSTM ()
input = get >>= \st -> put (st { inputs = True })

outstandingInput :: ST -> Bool
outstandingInput = (True==) . inputs

pushInput :: ST -> Int -> ST
pushInput st v = st { runstack = LitI v : runstack st }

clearInputs :: ST -> ST
clearInputs st = st { inputs = False }

clearOutputs :: ST -> ST
clearOutputs st = st { outputs = []}

output :: Literal -> String -> ExSTM ()
output l pf = do
  st <- get
  put (st { outputs = ((show l) ++ pf) : outputs st})


store :: String -> Literal -> Literal -> Literal -> ExSTM ()
store msg (LitI a) (LitI idx) v =
  do st0 <- get
     let fprog = program st0
         prog = take ((a+idx) + 1) fprog ++ [(Lit v,0)] ++ drop ((a+idx) + 2) fprog
     put (st0 { program = prog } )

store msg LitLocalsAddress (LitI idx) v = storeLocal idx v

store msg _ _ _ = throwError msg

load :: String -> Literal -> Literal -> ExSTM Literal
load msg (LitI a) (LitI idx) =
  do st0 <- get
     v <- check $ fst $ head $ drop ((a+idx) + 1) (program st0)
     return v
  where check (Lit v) = return v
        check _       = throwError msg

load msg LitArgsAddress (LitI idx) =
  do v <- loadArg idx
     return v

load msg LitLocalsAddress (LitI idx) =
  do v <- loadLocal idx
     return v

load msg _ _ = throwError msg

getProgram :: ExSTM [(Command,Line)]
getProgram = do st <- get
                return $ program st

lookupLabel :: String -> Name -> ExSTM Int
lookupLabel msg n = do st <- get
                       l <- findLabel' msg n (lmap st)
                       return l

findLabel' :: String -> Name -> LabelMap -> ExSTM Int
findLabel' msg n lmap =
  case M.lookup n lmap of
      Just a -> return a
      _      -> throwError msg

putRunStack :: RunStack -> ExSTM ()
putRunStack rs = do st0 <- get
                    put (st0 { runstack = rs } )

putRetStack :: ReturnStack -> ExSTM ()
putRetStack rs = do st0 <- get
                    put (st0 { retstack = rs } )

pushRunStack :: Literal -> ExSTM ()
pushRunStack v = do st0 <- get
                    putRunStack (v:runstack st0)

getRunStack :: ExSTM RunStack
getRunStack = get >>= return . runstack

pushRetStack :: Int -> ExSTM ()
pushRetStack v = do st0 <- get
                    putRetStack $ v:retstack st0

pop1Stack :: String -> (ST -> Stack a) -> (Stack a -> ExSTM ()) -> ExSTM a
pop1Stack msg sel p = do st <- get
                         let s = sel st
                         if (length s) == 0
                            then throwError msg -- stack is empty
                            else do p (tail s)
                                    return $ head s

pop1RunStack :: String -> ExSTM Literal
pop1RunStack msg = pop1Stack msg runstack putRunStack

pop1RetStack :: String -> ExSTM Int
pop1RetStack msg = pop1Stack msg retstack putRetStack

pop2Stack :: String -> (ST -> Stack a) -> (Stack a -> ExSTM ()) -> ExSTM (a,a)
pop2Stack msg sel p = do st <- get
                         let s = sel st
                         if (length s) < 2
                            then throwError msg
                            else do p (tail $ tail s)
                                    return $ (head s, head $ tail s)

pop2RunStack :: String -> ExSTM (Literal,Literal)
pop2RunStack msg = pop2Stack msg runstack putRunStack

pop2RetStack :: String -> ExSTM (Int,Int)
pop2RetStack msg = pop2Stack msg retstack putRetStack

incInstCount :: ExSTM ()
incInstCount = do st <- get
                  put (st { instCount = instCount st + 1 })

getLocals :: ExSTM Locals
getLocals = get >>= return . locals

getLocalsStack :: ExSTM LocalsStack
getLocalsStack = get >>= return . localsstack

popLocalsStack :: ExSTM ()
popLocalsStack =
  do st <- get
     put (st { localsstack = tail (localsstack st), locals = head (localsstack st) })

pushLocals :: ExSTM ()
pushLocals =
  do st <- get
     put (st { localsstack = (locals st) : localsstack st, locals = [] })

storeLocal :: Int -> Literal -> ExSTM ()
storeLocal n lit =
  do st <- get
     put (st { locals = (take n (locals st) ++ [lit] ++ drop (n+1) (locals st)) })

loadLocal :: Int -> ExSTM Literal
loadLocal n = getLocals >>= return . (!!n)

getCalleeArgs :: ExSTM Args
getCalleeArgs = get >>= return . calleeargs

getArgsStack :: ExSTM ArgsStack
getArgsStack = get >>= return . argsstack

popArgs :: ExSTM ()
popArgs =
  do st <- get
     put (st { argsstack = tail (argsstack st),
               calleeargs = head (argsstack st),
               callerargs = [] })

pushArgs :: ExSTM ()
pushArgs =
  do st <- get
     put (st { argsstack = (calleeargs st) : argsstack st,
               calleeargs = reverse $ callerargs st,
               callerargs = [] })

loadArg :: Int -> ExSTM Literal
loadArg n = getCalleeArgs >>= return . (!!n)

pushArg :: Literal -> ExSTM ()
pushArg lit =
  do st <- get
     put (st { callerargs = lit : callerargs st })

getInstCount :: ExSTM Int
getInstCount = get >>= return . instCount

getRetReg :: String -> ExSTM Literal
getRetReg msg =
  do st <- get
     case returnReg st of
          Just l -> return l
          Nothing -> throwError msg

putRetReg :: Maybe Literal -> ExSTM ()
putRetReg lit =
  do st <- get
     put (st { returnReg = lit })

------------------------------------------------------------
-- Some helper functions
------------------------------------------------------------

unlit1 (LitI a) _   = return a
unlit1 _        msg = throwError msg

compareOp :: String -> CompOp -> Literal -> Literal ->ExSTM Int
compareOp _ op (LitI i) (LitI i') = return $ fromBool $ compareOp' op
  where compareOp' CEq  = i == i'
        compareOp' CNEq = i /= i'
        compareOp' CLEq = i <= i'
        compareOp' CGEq = i >= i'
        compareOp' CLT  = i < i'
        compareOp' CGT  = i > i'

compareOp msg _ _ _ = throwError msg

arithBinaryOp :: String ->
                   (Int -> Int -> Int) ->
                     Literal ->
                       Literal ->
                         ExSTM Int
arithBinaryOp _ op (LitI i) (LitI i') = return $ op i i'
arithBinaryOp msg _ _ _ = throwError msg

arithUnaryOp :: String ->
                   (Int -> Int) ->
                     Literal ->
                       ExSTM Int
arithUnaryOp _ op (LitI i) = return $ op i
arithUnaryOp msg _ _ = throwError msg

fromBool :: Bool -> Int
fromBool True  = 0
fromBool False = 1

----------------------------------------------------
-- Monad Based Evaluator
----------------------------------------------------

begin :: Program -> IO (Either String (ST, [(Command,Line)]))
begin (Program p lmap) =
  runExceptT $ evalStateT (do
    l <- findLabel' "main not defined"
                    "main"
                    lmap
    incInstCount
    (s, cmds) <- eval'' step'' p
    return (s, cmds)) (ST p lmap [] [] [] [] [] [] [] Nothing 0 [] False)

step :: ST -> [(Command,Line)] -> IO (Either String (ST, [(Command,Line)]))
step s p =
  runExceptT $ evalStateT (do
    incInstCount
    (s, cmds) <- eval'' step'' p
    return (s, cmds)) s

step'' :: [(Command,Line)] -> ExSTM (ST, [(Command,Line)])
step'' cmds = get >>= \s -> return (s, cmds)

eval'' :: ([(Command,Line)] ->
            ExSTM (ST,[(Command,Line)])) ->
              [(Command,Line)] ->
                ExSTM (ST,[(Command,Line)])

eval'' cond [] =  throwError "unexpected end of program, i.e. no HALT"

eval'' cond ((Nop,l):rem)      = cond rem

eval'' cond ((Halt,l):_)       = cond []

eval'' cond ((Push lit,l):rem) = pushRunStack lit >> cond rem

eval'' cond ((Pop,l):rem)      =
  pop1RunStack "tried to POP from empty stack" >>  cond rem

eval'' cond ((DumpStack,l):rem) =
  do  rs <- getRunStack
      liftIO $ putStr $ (show rs ++ "\n")
      cond rem

eval'' cond ((Label _ _,l):rem) = cond rem

eval'' cond ((Out,l):rem) =
  do v <- pop1RunStack "tried OUT from empty stack"
     output v ""
     cond rem

eval'' cond ((OutLN,l):rem) =
  do v <- pop1RunStack "tried OUTLN from empty stack"
     output v "\n"
     cond rem

eval'' cond ((In,l):rem) = input >> cond rem

eval'' cond ((Flush,l):rem) =
  do liftIO $ hFlush stdout
     cond rem

eval'' cond ((PushArg,l):rem) =
  do v <- pop1RunStack "tried PUSHARG from empty stack"
     pushArg v
     cond rem

eval'' cond ((MovRet,l):rem) =
  do v <- pop1RunStack "tried MOVRET from empty stack"
     putRetReg $ Just v
     cond rem

eval'' cond ((PushRet,l):rem) =
  do v <- getRetReg "tried to read undefined return reg"
     pushRunStack v
     cond rem

eval'' cond ((Dup,l):rem) =
  do v <- pop1RunStack "DUP expects one int on the stack"
     pushRunStack v
     pushRunStack v
     cond rem

eval'' cond ((Swap,l):rem) =
  do (a,b) <- pop2RunStack "SWAP expects two values on the stack"
     pushRunStack a
     pushRunStack b
     cond rem

eval'' cond ((Comp op,l):rem) =
  do let msg = "COMP expexts two ints on the stack"
     (a,b) <- pop2RunStack msg
     res <- compareOp msg op a b
     pushRunStack $ LitI res
     cond rem

eval'' cond ((Compl,l):rem) =
  do let msg = "COMPL expexts one int on the stack"
     a <- pop1RunStack msg
     res <- compareOp msg CEq a (LitI 1)
     pushRunStack $ LitI res
     cond rem

eval'' cond ((Add,l):rem) =
  do let msg = "ADD expects two ints on the stack"
     (a,b) <- pop2RunStack msg
     res <- arithBinaryOp msg (+) a b
     pushRunStack $ LitI res
     cond rem

eval'' cond ((Minus,l):rem) =
  do let msg = "MINUS expects two ints on the stack"
     (a,b) <- pop2RunStack msg
     res <- arithBinaryOp msg (-) a b
     pushRunStack $ LitI res
     cond rem

eval'' cond ((Mul,l):rem) =
  do let msg = "Mul expects two ints on the stack"
     (a,b) <- pop2RunStack msg
     res <- arithBinaryOp msg (*) a b
     pushRunStack $ LitI res
     cond rem

eval'' cond ((Div,l):rem) =
  do let msg = "Div expects two ints on the stack"
     (a,b) <- pop2RunStack msg
     res <- arithBinaryOp msg (quot) a b
     pushRunStack $ LitI res
     cond rem

eval'' cond ((And,l):rem) =
  do let msg = "AND expects two ints on the stack"
     (a,b) <- pop2RunStack msg
     res <- arithBinaryOp msg (.&.) a b
     pushRunStack $ LitI res
     cond rem

eval'' cond ((Or,l):rem) =
  do let msg = "AND expects two ints on the stack"
     (a,b) <- pop2RunStack msg
     res <- arithBinaryOp msg (.|.) a b
     pushRunStack $ LitI res
     cond rem

eval'' cond ((Not,l):rem) =
  do let msg = "NOT expects one int on the stack"
     a <- pop1RunStack msg
     res <- arithUnaryOp msg complement a
     pushRunStack $ LitI res
     cond rem

eval'' cond ((JmpNZ,l):rem) =
  do let msg = "JMPNZ expects one int and an address on the stack"
     (addr, a) <- pop2RunStack msg
     if a == LitI 0
        then cond rem
        else do addr' <- unlit1 addr msg
                prog <- getProgram
                cond $ drop addr' prog

eval'' cond ((Jmp,l):_) =
  do let msg = "JMP expects an address on the stack"
     a <- pop1RunStack msg
     a' <- unlit1 a msg
     prog <- getProgram
     cond $ drop a' prog

eval'' cond ((Subcall,l):rem) =
  do let msg = "SUBCALL expects an address on the stack"
     a <- pop1RunStack msg
     a' <- unlit1 a msg
     prog <- getProgram
     pushRetStack ((length prog - length rem))
     pushArgs
     cond $ drop a' prog

eval'' cond ((Return,l):_) =
  do let msg = "RETURN expects an address on the return stack"
     a <- pop1RetStack msg
     popArgs
     prog <- getProgram
     cond (drop a prog)

eval'' cond ((AddrOf n,l):rem) =
  do let msg = "Label " ++ n ++ " not found in ADDROF"
     l <- lookupLabel msg n
     pushRunStack $ LitI l
     cond rem

eval'' cond ((Store,l):rem) =
  do let msg = "STORE expects one int and an address on the stack"
     (b,addr) <- pop2RunStack msg
     store msg addr (LitI 0) b
     cond rem

eval'' cond ((StoreIdx,l):rem) =
  do let msg = "STORE expects one int and an address on the stack"
     (idx,addr) <- pop2RunStack msg
     b          <- pop1RunStack msg
     store msg addr idx b
     cond rem

eval'' cond ((Load,l):rem) =
  do let msg = "LOAD expects an address on the stack"
     a <- pop1RunStack msg
     v <- load msg a (LitI 0)
     pushRunStack v
     cond rem

eval'' cond ((LoadIdx,l):rem) =
  do let msg = "LOAD expects an address on the stack"
     (idx, a) <- pop2RunStack msg
     v <- load msg a idx
     pushRunStack v
     cond rem

eval'' cond ((cmd,_):rem) =
  throwError ("unimplemented instruction " ++ show cmd)
